from dataclasses import dataclass
import operator

from .seqs import kscore
from .util import sliding_sum


@dataclass(slots=True)
class Read:
    desc: str
    seq: str
    qual: str


def seq_id(read):
    return read.desc.split()[0]


def length(read):
    return len(read.seq)


def qvals(read, offset=33):
    return [ord(x) - offset for x in read.qual]


def trim(read, start_idx=0, end_idx=100):
    seq = read.seq[start_idx:end_idx]
    qual = read.qual[start_idx:end_idx]
    return Read(read.desc, seq, qual)


def kscore_ok(read, k=4, min_kscore=0.55):
    return kscore(read.seq, k=k) >= min_kscore


def length_ok(read, threshold=100, cmp=operator.ge):
    return cmp(length(read), threshold)


def seq_id_ok(read, seq_ids, keep=False):
    id = seq_id(read)
    if keep:
        return id in seq_ids
    else:
        return id not in seq_ids


def trim_moving_average(read, k=4, threshold=15):
    window_sum_threshold = threshold * k
    qs = qvals(read)
    for window_idx, window_sum in enumerate(sliding_sum(qs, k=k)):
        if window_sum < window_sum_threshold:
            end_idx = window_idx
            # Extend to include last qval in window meeting threshold
            for extended_idx in range(window_idx, window_idx + k):
                if qs[extended_idx] >= threshold:
                    end_idx = extended_idx + 1
            return trim(read, end_idx=end_idx)
    return read


def trim_ends(read, threshold_start=3, threshold_end=3):
    qs = qvals(read)
    trim_start = 0
    for i, q in enumerate(qs):
        if q < threshold_start:
            trim_start = i + 1
        else:
            break
    trim_end = 0
    for i, q in enumerate(reversed(qs)):
        if q < threshold_end:
            trim_end = i + 1
        else:
            break
    if (trim_start == 0) and (trim_end == 0):
        return read
    else:
        start_idx = trim_start
        end_idx = length(read) - trim_end
        return trim(read, start_idx=start_idx, end_idx=end_idx)



import random

from heyfastqlib import util


def test_subsample():
    xs = ["a", "b", "c", "d", "e", "f"]
    # sampling less than or equal to n is original list
    assert util.subsample(xs[:4], n=4) == xs[:4]
    # if any of the first n elements are in the resulting list,
    # they should be in their original position
    assert util.subsample(xs, n=3, seed=0) == ["f", "b", "c"]


def test_sliding_sum():
    xs = [1, 2, 3, 4, 5]
    assert list(util.sliding_sum(xs, 3)) == [6, 9, 12]



import gzip
from heyfastqlib.command import *


def test_command(tmp_path):
    in1 = tmp_path / "input_1.fastq"
    with open(in1, "w") as f:
        f.write("@a\nCGTT\n+\n;=GG\n@b\nACTG\n+\nGGGG\n")
    in2 = tmp_path / "input_2.fastq"
    with open(in2, "w") as f:
        f.write("@a\nAACG\n+\n;=GG\n@b\nCAGT\n+\nGGGG\n")
    out1 = tmp_path / "output_1.fastq"
    out2 = tmp_path / "output_2.fastq"
    heyfastq_main(
        [
            "trim-fixed",
            "--length",
            "2",
            "--input",
            str(in1),
            str(in2),
            "--output",
            str(out1),
            str(out2),
        ]
    )
    with open(out1) as f:
        assert f.read() == "@a\nCG\n+\n;=\n@b\nAC\n+\nGG\n"
    with open(out2) as f:
        assert f.read() == "@a\nAA\n+\n;=\n@b\nCA\n+\nGG\n"


def test_gzip_command(tmp_path):
    in1 = tmp_path / "input_1.fastq.gz"
    with gzip.open(in1, "wt") as f:
        f.write("@a\nCGTT\n+\n;=GG\n@b\nACTG\n+\nGGGG\n")
    in2 = tmp_path / "input_2.fastq.gz"
    with gzip.open(in2, "wt") as f:
        f.write("@a\nAACG\n+\n;=GG\n@b\nCAGT\n+\nGGGG\n")
    out1 = tmp_path / "output_1.fastq.gz"
    out2 = tmp_path / "output_2.fastq.gz"
    heyfastq_main(
        [
            "trim-fixed",
            "--length",
            "2",
            "--input",
            str(in1),
            str(in2),
            "--output",
            str(out1),
            str(out2),
        ]
    )
    with gzip.open(out1, "rt") as f:
        assert f.read() == "@a\nCG\n+\n;=\n@b\nAC\n+\nGG\n"
    with gzip.open(out2, "rt") as f:
        assert f.read() == "@a\nAA\n+\n;=\n@b\nCA\n+\nGG\n"


def test_trim_qual_command(tmp_path):
    in1 = tmp_path / "input_1.fastq"
    with open(in1, "w") as f:
        f.write("@a\nAACGTACGTGGGGGGGG\n+\n&55555555&&&&&&&&")
    in2 = tmp_path / "input_2.fastq"
    with open(in2, "w") as f:
        f.write("@a\nCGTTCGTTAAAAAAAA\n+\n55555555!!!!!!!!")
    out1 = tmp_path / "output_1.fastq"
    out2 = tmp_path / "output_2.fastq"
    heyfastq_main(
        [
            "trim-qual",
            "--window-width",
            "4",
            "--window-threshold",
            "7",
            "--start-threshold",
            "6",
            "--min-length",
            "4",
            "--input",
            str(in1),
            str(in2),
            "--output",
            str(out1),
            str(out2),
        ]
    )
    with open(out1) as f:
        # Final Q20 base passes
        # mean([20, 5, 5, 5]) = 8.75
        assert f.read() == "@a\nACGTACGT\n+\n55555555\n"
    with open(out2) as f:
        # Final Q20 base fails, but is added in extension
        # mean([20, 0, 0, 0]) = 5
        assert f.read() == "@a\nCGTTCGTT\n+\n55555555\n"


in1_kscore = """\
@a
AAAATAAAAAAAAAA
+
===============
@b
GCTAGCTAGCATGCATCTA
+
===================
"""

in2_kscore = """\
@a
GCTACGATCAGTACGAT
+
=================
@b
GCTGAGCTACGGTC
+
==============
"""


def test_filter_kscore_command(tmp_path):
    in1 = tmp_path / "input_1.fastq"
    with open(in1, "w") as f:
        f.write(in1_kscore)
    in2 = tmp_path / "input_2.fastq"
    with open(in2, "w") as f:
        f.write(in2_kscore)
    out1 = tmp_path / "output_1.fastq"
    out2 = tmp_path / "output_2.fastq"
    heyfastq_main(
        [
            "filter-kscore",
            "--min-kscore",
            "0.55",
            "--input",
            str(in1),
            str(in2),
            "--output",
            str(out1),
            str(out2),
        ]
    )
    with open(out1) as f:
        assert f.read() == "@b\nGCTAGCTAGCATGCATCTA\n+\n===================\n"
    with open(out2) as f:
        assert f.read() == "@b\nGCTGAGCTACGGTC\n+\n==============\n"


in1_length = """\
@a
ACGTACGTACGT
+
123456789012
@b
AAGGC
+
12345
"""

in2_length = """\
@a
AGGTCGTCTAAC
+
123456789012
@b
AGCTGCTACGCTA
+
1234567890123
"""


def test_filter_length_command(tmp_path):
    in1 = tmp_path / "input_1.fastq"
    with open(in1, "w") as f:
        f.write(in1_length)
    in2 = tmp_path / "input_2.fastq"
    with open(in2, "w") as f:
        f.write(in2_length)
    out1 = tmp_path / "output_1.fastq"
    out2 = tmp_path / "output_2.fastq"
    heyfastq_main(
        [
            "filter-length",
            "--length",
            "6",
            "--input",
            str(in1),
            str(in2),
            "--output",
            str(out1),
            str(out2),
        ]
    )
    with open(out1) as f:
        assert f.read() == "@a\nACGTACGTACGT\n+\n123456789012\n"
    with open(out2) as f:
        assert f.read() == "@a\nAGGTCGTCTAAC\n+\n123456789012\n"


in1_seqids = """\
@a
CGTA
+
1234
@b
GTCC
+
5678
"""


def test_filter_seq_ids_command(tmp_path):
    in1 = tmp_path / "input_1.fastq"
    with open(in1, "w") as f:
        f.write(in1_seqids)
    ids_path = tmp_path / "ids.txt"
    with open(ids_path, "w") as f:
        f.write("a\n")
    out1 = tmp_path / "output_1.fastq"
    heyfastq_main(
        [
            "filter-seqids",
            str(ids_path),
            "--input",
            str(in1),
            "--output",
            str(out1),
        ]
    )
    with open(out1) as f:
        assert f.read() == "@b\nGTCC\n+\n5678\n"


in1_subsample = """\
@a
AGC
+
123
@b
GCT
+
.o.
@c
CTG
+
***
"""

in2_subsample = """\
@a
ACG
+
998
@b
GCT
+
..o
@c
CTC
+
xxx
"""


def test_subsample_command(tmp_path):
    in1 = tmp_path / "input_1.fastq"
    with open(in1, "w") as f:
        f.write(in1_subsample)
    in2 = tmp_path / "input_2.fastq"
    with open(in2, "w") as f:
        f.write(in2_subsample)
    out1 = tmp_path / "output_1.fastq"
    out2 = tmp_path / "output_2.fastq"
    heyfastq_main(
        [
            "subsample",
            "--n",
            "2",
            "--seed",
            "500",
            "--input",
            str(in1),
            str(in2),
            "--output",
            str(out1),
            str(out2),
        ]
    )
    with open(out1) as f:
        assert f.read() == "@a\nAGC\n+\n123\n@c\nCTG\n+\n***\n"



from heyfastqlib.seqs import *


def test_kmers():
    seq = "ATGCGCT"
    assert list(kmers(seq, k=4)) == ["ATGC", "TGCG", "GCGC", "CGCT"]
    assert list(kmers(seq, k=5)) == ["ATGCG", "TGCGC", "GCGCT"]


def test_kscore():
    assert kscore("AAAAA", k=4) == 1 / 5
    assert kscore("AAAATAAAAT", k=4) == 5 / 10



def map_paired(paired_reads, f, **kwargs):
    for paired_read in paired_reads:
        yield tuple(f(read, **kwargs) for read in paired_read)


def filter_paired(paired_reads, f, requirement=all, **kwargs):
    for paired_read in paired_reads:
        passes_filter = (f(read, **kwargs) for read in paired_read)
        if requirement(passes_filter):
            yield paired_read



import collections
import itertools
import random


def subsample(xs, n, seed=None):
    random.seed(seed)
    # https://en.wikipedia.org/wiki/Reservoir_sampling
    reservoir = [None for _ in range(n)]
    for i, x in enumerate(xs):
        if i < n:
            reservoir[i] = x
        else:
            idx = random.randint(0, i)
            if idx < n:
                reservoir[idx] = x
    return reservoir


def sliding_sum(xs, k=4):
    # From moving_average recipe in Python docs
    xs = iter(xs)
    d = collections.deque(itertools.islice(xs, k - 1))
    d.appendleft(0)
    s = sum(d)
    for elem in xs:
        s += elem - d.popleft()
        d.append(elem)
        yield s



from .read import Read


def _grouper(iterable, n):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3) --> ABC DEF
    args = [iter(iterable)] * n
    return zip(*args)


def parse_fastq(f):
    for desc, seq, _, qual in _grouper(f, 4):
        desc = desc.rstrip()[1:]
        seq = seq.rstrip()
        qual = qual.rstrip()
        yield Read(desc, seq, qual)


def parse_fastq_paired(fs):
    fastq_iters = [parse_fastq(f) for f in fs]
    for paired_rec in zip(*fastq_iters):
        yield paired_rec


def write_fastq(f, reads):
    for read in reads:
        f.write("@{0}\n{1}\n+\n{2}\n".format(read.desc, read.seq, read.qual))


def write_fastq_paired(fs, paired_recs):
    fs = list(fs)
    for paired_rec in paired_recs:
        for f, read in zip(fs, paired_rec):
            f.write("@{0}\n{1}\n+\n{2}\n".format(read.desc, read.seq, read.qual))


def parse_seq_ids(f):
    for line in f:
        line = line.strip()
        if line.startswith("#") or (line == ""):
            continue
        seq_id = line.split()[0]
        yield seq_id



import gzip
from heyfastqlib.argparse_types import GzipFileType


def test_gzipfiletype_init(tmp_path):
    gzft = GzipFileType()
    assert gzft._mode == "r"


def test_gzipfiletype_call(tmp_path):
    gzft = GzipFileType()
    with open(tmp_path / "test_in.txt", "w") as f:
        f.write("test")
    with gzft(str(tmp_path / "test_in.txt")) as f:
        assert f.read() == "test"

    gzftw = GzipFileType(mode="w")
    with gzftw(str(tmp_path / "test_out.txt")) as f:
        f.write("test")
    with gzft(str(tmp_path / "test_out.txt")) as f:
        assert f.read() == "test"


def test_gzipfiletype_call_gz(tmp_path):
    gzft = GzipFileType()
    with gzip.open(tmp_path / "test_in.txt.gz", "wt") as f:
        f.write("test")
    with gzft(str(tmp_path / "test_in.txt.gz")) as f:
        assert f.read() == "test"

    gzftw = GzipFileType(mode="wt")
    with gzftw(str(tmp_path / "test_out.txt.gz")) as f:
        f.write("test")
    with gzft(str(tmp_path / "test_out.txt.gz")) as f:
        assert f.read() == "test"



from heyfastqlib.read import Read, length, length_ok
from heyfastqlib.paired_reads import *

reads = [
    (Read("a", "AAAG", "1234"), Read("a", "CTTT", "1234")),
    (Read("b", "ACGTACGT", "12345678"), Read("b", "ACG", "123")),
]


def test_map_paired():
    assert list(map_paired(reads, length)) == [(4, 4), (8, 3)]


def test_filter_paired():
    assert list(filter_paired(reads, length_ok, threshold=4)) == reads[:1]

    assert list(filter_paired(reads, length_ok, requirement=any, threshold=4)) == reads



from heyfastqlib.io import *


def test_parse_fastq():
    f = [
        "@ab c",
        "GGCA",
        "+",
        "==;G",
        "@d:e:f",
        "CCGT",
        "+",
        "1,4E",
    ]
    reads = parse_fastq(f)
    assert list(reads) == [
        Read("ab c", "GGCA", "==;G"),
        Read("d:e:f", "CCGT", "1,4E"),
    ]


def test_parse_fastq_paired():
    fq1 = ["@a", "TA", "+", "GG", "@b", "CG", "+", "AB"]
    fq2 = ["@a", "AG", "+", "FF", "@b", "TC", "+", "BC"]
    recs = parse_fastq_paired((fq1, fq2))
    assert list(recs) == [
        (Read("a", "TA", "GG"), Read("a", "AG", "FF")),
        (Read("b", "CG", "AB"), Read("b", "TC", "BC")),
    ]


class MockFile:
    def __init__(self):
        self.contents = ""

    def write(self, x):
        self.contents += x


def test_write_fastq():
    f = MockFile()
    reads = [Read("a", "CGT", "BBC"), Read("b", "TAC", "CCD")]
    write_fastq(f, reads)
    assert f.contents == "@a\nCGT\n+\nBBC\n@b\nTAC\n+\nCCD\n"


def test_write_fastq_paired():
    f1 = MockFile()
    f2 = MockFile()
    paired_recs = [
        (Read("a", "CGT", "BBC"), Read("a", "ACG", "CCD")),
        (Read("b", "GTA", "AAB"), Read("b", "TAC", "EEF")),
    ]
    write_fastq_paired((f1, f2), paired_recs)
    assert f1.contents == "@a\nCGT\n+\nBBC\n@b\nGTA\n+\nAAB\n"
    assert f2.contents == "@a\nACG\n+\nCCD\n@b\nTAC\n+\nEEF\n"


def test_parse_seq_ids():
    f = ["Id1\n", "\tId2|345 678  \n", "   ", "   # a comment", "  id3"]
    assert list(parse_seq_ids(f)) == ["Id1", "Id2|345", "id3"]



from types import SimpleNamespace
import io

import pytest

from heyfastqlib.command import (
    ReadStats,
    _iter_with_stats,
    _run_paired_command,
)
from heyfastqlib.paired_reads import map_paired
from heyfastqlib.read import Read


def _make_read(desc: str, seq: str) -> Read:
    return Read(desc, seq, "I" * len(seq))


def _make_fastq_text(reads):
    return "".join(f"@{read.desc}\n{read.seq}\n+\n{read.qual}\n" for read in reads)


def test_read_stats_observes_paired_reads():
    stats = ReadStats()
    paired_read = (
        _make_read("r1/1", "AAAA"),
        _make_read("r1/2", "AAA"),
    )

    stats.observe(paired_read)

    assert stats.total_reads == 2
    assert stats.total_bases == 7
    assert pytest.approx(stats.average_length) == 3.5


def test_iter_with_stats_passthroughs_reads():
    stats = ReadStats()
    pairs = [
        (
            _make_read("r1/1", "AAAA"),
            _make_read("r1/2", "AAA"),
        ),
        (
            _make_read("r2/1", "CC"),
            _make_read("r2/2", "C"),
        ),
    ]

    result = list(_iter_with_stats(iter(pairs), stats))

    assert result == pairs
    assert stats.total_reads == 4
    assert stats.total_bases == 10
    assert pytest.approx(stats.average_length) == 2.5


def test_run_paired_command_logs_input_and_output_stats(capsys):
    mate1_reads = [
        _make_read("r1/1", "AAAA"),
        _make_read("r2/1", "CCCCC"),
    ]
    mate2_reads = [
        _make_read("r1/2", "GGG"),
        _make_read("r2/2", "TTTTTT"),
    ]

    def shorten(read: Read) -> Read:
        return Read(read.desc, read.seq[:-1], read.qual[:-1])

    def transform(reads):
        return map_paired(reads, shorten)

    args = SimpleNamespace(
        input=[
            io.StringIO(_make_fastq_text(mate1_reads)),
            io.StringIO(_make_fastq_text(mate2_reads)),
        ],
        output=[io.StringIO(), io.StringIO()],
    )

    _run_paired_command(args, transform)

    stderr = capsys.readouterr().err.strip().splitlines()
    assert stderr[0] == "Input reads: total=4, average_length=4.50"
    assert stderr[1] == "Output reads: total=4, average_length=3.50"

    output_reads = [
        _make_fastq_text([Read(r.desc, r.seq[:-1], r.qual[:-1]) for r in mate1_reads]),
        _make_fastq_text([Read(r.desc, r.seq[:-1], r.qual[:-1]) for r in mate2_reads]),
    ]

    assert args.output[0].getvalue() == output_reads[0]
    assert args.output[1].getvalue() == output_reads[1]



import gzip
import sys
import argparse


class GzipFileType(object):
    """Factory for creating optionally gzipped file object types

    Taken from https://github.com/python/cpython/blob/05a370abd6cdfe4b54be60b3b911f3a441026bb2/Lib/argparse.py

    Instances of GzipFileType are typically passed as type= arguments to the
    ArgumentParser add_argument() method.

    Keyword Arguments:
        - mode -- A string indicating how the file is to be opened. Accepts the
            same values as the builtin open() function.
        - bufsize -- The file's desired buffer size. Accepts the same values as
            the builtin open() function.
        - encoding -- The file's encoding. Accepts the same values as the
            builtin open() function.
        - errors -- A string indicating how encoding and decoding errors are to
            be handled. Accepts the same value as the builtin open() function.
    """

    def __init__(self, mode="r", bufsize=-1, encoding=None, errors=None):
        self._mode = mode
        self._bufsize = bufsize
        self._encoding = encoding
        self._errors = errors

    def __call__(self, string):
        # the special argument "-" means sys.std{in,out}
        if string == "-":
            if "r" in self._mode:
                return sys.stdin.buffer if "b" in self._mode else sys.stdin
            elif any(c in self._mode for c in "wax"):
                return sys.stdout.buffer if "b" in self._mode else sys.stdout
            else:
                msg = f'argument "-" with mode {self.mode}'
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            try:
                with open(string, "rb") as test_f:
                    gzipped = test_f.read(2) == b"\x1f\x8b"
            except FileNotFoundError:
                gzipped = string.endswith(".gz")

            if gzipped:
                f = gzip.open(
                    string,
                    f"{self._mode}t",
                    self._bufsize,
                    self._encoding,
                    self._errors,
                )
            else:
                f = open(
                    string, self._mode, self._bufsize, self._encoding, self._errors
                )

            return f
        except OSError as e:
            args = {"filename": string, "error": e}
            message = f"can't open {args['filename']}: {args['error']}"
            raise argparse.ArgumentTypeError(message % args)

    def __repr__(self):
        args = self._mode, self._bufsize
        kwargs = [("encoding", self._encoding), ("errors", self._errors)]
        args_str = ", ".join(
            [repr(arg) for arg in args if arg != -1]
            + ["%s=%r" % (kw, arg) for kw, arg in kwargs if arg is not None]
        )
        return "%s(%s)" % (type(self).__name__, args_str)


class HFQFormatter(argparse.HelpFormatter):
    def __init__(self, prog, indent_increment=2, max_help_position=25, width=None):
        super().__init__(prog, indent_increment, max_help_position, width)

    # based on ArgumentDefaultsHelpFormatter but with a different search string
    def _get_help_string(self, action):
        help = action.help
        if help is None:
            help = ""

        if "default" not in help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += " (default: %(default)s)"
        return help



import argparse
import operator
import signal
import sys
from dataclasses import dataclass

from . import __version__
from .util import (
    subsample,
)
from .io import (
    parse_fastq_paired,
    write_fastq_paired,
    parse_seq_ids,
)
from .paired_reads import map_paired, filter_paired
from .read import (
    trim,
    kscore_ok,
    length_ok,
    seq_id_ok,
    trim_moving_average,
    trim_ends,
)
from .argparse_types import GzipFileType, HFQFormatter


@dataclass
class ReadStats:
    total_reads: int = 0
    total_bases: int = 0

    def observe(self, paired_read):
        for read in paired_read:
            self.total_reads += 1
            self.total_bases += len(read.seq)

    @property
    def average_length(self):
        if self.total_reads == 0:
            return 0.0
        return self.total_bases / self.total_reads


def _iter_with_stats(paired_reads, stats):
    for paired_read in paired_reads:
        stats.observe(paired_read)
        yield paired_read


def _log_stats(label, stats):
    print(
        f"{label} reads: total={stats.total_reads}, average_length={stats.average_length:.2f}",
        file=sys.stderr,
    )


def _run_paired_command(args, build_output):
    input_stats = ReadStats()
    output_stats = ReadStats()

    reads = _iter_with_stats(parse_fastq_paired(args.input), input_stats)
    out_reads = build_output(reads)
    write_fastq_paired(args.output, _iter_with_stats(out_reads, output_stats))

    _log_stats("Input", input_stats)
    _log_stats("Output", output_stats)


def subsample_subcommand(args):
    def transform(reads):
        return subsample(reads, args.n, args.seed)

    _run_paired_command(
        args,
        transform,
    )


def trim_fixed_subcommand(args):
    def transform(reads):
        return map_paired(reads, trim, end_idx=args.length)

    _run_paired_command(
        args,
        transform,
    )


def trim_qual_subcommand(args):
    def transform(reads):
        return _trim_quality_pipeline(
            reads,
            window_width=args.window_width,
            window_threshold=args.window_threshold,
            start_threshold=args.start_threshold,
            end_threshold=args.end_threshold,
            min_length=args.min_length,
        )

    _run_paired_command(
        args,
        transform,
    )


def filter_length_subcommand(args):
    cmp = operator.lt if args.less else operator.ge

    def transform(reads):
        return filter_paired(reads, length_ok, threshold=args.length, cmp=cmp)

    _run_paired_command(
        args,
        transform,
    )


def filter_kscore_subcommand(args):
    def transform(reads):
        return filter_paired(
            reads, kscore_ok, k=args.kmer_size, min_kscore=args.min_kscore
        )

    _run_paired_command(
        args,
        transform,
    )


def filter_seq_ids_subcommand(args):
    seq_ids = set(parse_seq_ids(args.idsfile))

    def transform(reads):
        return filter_paired(reads, seq_id_ok, seq_ids=seq_ids, keep=args.keep_ids)

    _run_paired_command(
        args,
        transform,
    )


def _trim_quality_pipeline(
    reads,
    *,
    window_width,
    window_threshold,
    start_threshold,
    end_threshold,
    min_length,
):
    trimmed_moving_average_reads = map_paired(
        reads, trim_moving_average, k=window_width, threshold=window_threshold
    )
    trimmed_ends_reads = map_paired(
        trimmed_moving_average_reads,
        trim_ends,
        threshold_start=start_threshold,
        threshold_end=end_threshold,
    )
    return filter_paired(trimmed_ends_reads, length_ok, threshold=min_length)


fastq_io_parser = argparse.ArgumentParser(add_help=False, formatter_class=HFQFormatter)
fastq_io_parser.add_argument(
    "--input",
    type=GzipFileType("r"),
    nargs="*",
    default=[sys.stdin],
    help="Input FASTQs, can be gzipped (default: stdin)",
)
fastq_io_parser.add_argument(
    "--output",
    type=GzipFileType("w"),
    nargs="*",
    default=[sys.stdout],
    help="Output FASTQs, can be gzipped (default: stdout)",
)


def heyfastq_main(argv=None):
    # Ignore SIG_PIPE and don't throw exceptions on it
    # newbebweb.blogspot.com/2012/02/python-head-ioerror-errno-32-broken.html
    # Try/catch to not fail on Windows
    # https://github.com/t2mune/mrtparse/issues/18
    try:
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except AttributeError:
        pass

    main_parser = argparse.ArgumentParser()
    main_parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{__version__}",
    )
    subparsers = main_parser.add_subparsers(title="Subcommands", required=True)

    trim_fixed_parser = subparsers.add_parser(
        "trim-fixed",
        parents=[fastq_io_parser],
        formatter_class=HFQFormatter,
        help="Trim reads to fixed length",
    )
    trim_fixed_parser.add_argument(
        "--length",
        type=int,
        default=100,
        help="Length of output reads",
    )
    trim_fixed_parser.set_defaults(func=trim_fixed_subcommand)

    trim_qual_parser = subparsers.add_parser(
        "trim-qual",
        parents=[fastq_io_parser],
        formatter_class=HFQFormatter,
        help="Trim reads based on quality scores",
    )
    trim_qual_parser.add_argument(
        "--window-width", type=int, default=4, help="Sliding window width"
    )
    trim_qual_parser.add_argument(
        "--window-threshold",
        type=float,
        default=15,
        help="Sliding window mean quality threshold",
    )
    trim_qual_parser.add_argument(
        "--start-threshold",
        type=float,
        default=3,
        help="Quality threshold for trimming start of read",
    )
    trim_qual_parser.add_argument(
        "--end-threshold",
        type=float,
        default=3,
        help="Quality threshold for trimming end of read",
    )
    trim_qual_parser.add_argument(
        "--min-length",
        type=int,
        default=36,
        help="Minimum length after quality trimming",
    )
    trim_qual_parser.set_defaults(func=trim_qual_subcommand)

    filter_length_parser = subparsers.add_parser(
        "filter-length",
        parents=[fastq_io_parser],
        formatter_class=HFQFormatter,
        help="Filter reads by length",
    )
    filter_length_parser.add_argument(
        "--length",
        type=int,
        default=100,
        help="Length threshold",
    )
    filter_length_parser.add_argument(
        "--less",
        action="store_true",
        help=(
            "Keep reads that are less than the specified length "
            "(default: keep greater than or equal to length)"
        ),
    )
    filter_length_parser.set_defaults(func=filter_length_subcommand)

    filter_kscore_parser = subparsers.add_parser(
        "filter-kscore",
        parents=[fastq_io_parser],
        formatter_class=HFQFormatter,
        help="Filter reads by komplexity score",
    )
    filter_kscore_parser.add_argument(
        "--kmer-size", type=int, default=4, help="Kmer size"
    )
    filter_kscore_parser.add_argument(
        "--min-kscore",
        type=float,
        default=0.55,
        help="Minimum komplexity score",
    )
    filter_kscore_parser.set_defaults(func=filter_kscore_subcommand)

    filter_seq_ids_parser = subparsers.add_parser(
        "filter-seqids",
        parents=[fastq_io_parser],
        formatter_class=HFQFormatter,
        help="Filter reads by sequence id",
    )
    filter_seq_ids_parser.add_argument(
        "idsfile",
        type=argparse.FileType("r"),
        help="File containing sequence ids, one per line",
    )
    filter_seq_ids_parser.add_argument(
        "--keep-ids",
        action="store_true",
        help="Keep, rather than remove ids in list",
    )
    filter_seq_ids_parser.set_defaults(func=filter_seq_ids_subcommand)

    subsample_parser = subparsers.add_parser(
        "subsample",
        parents=[fastq_io_parser],
        formatter_class=HFQFormatter,
        help="Select random reads",
    )
    subsample_parser.add_argument("--n", type=int, default=1000, help="Number of reads")
    subsample_parser.add_argument("--seed", type=int, help="Random seed")
    subsample_parser.set_defaults(func=subsample_subcommand)

    args = main_parser.parse_args(argv)
    if args.input is None:  # pragma: no cover
        args.input = sys.stdin
    if args.output is None:  # pragma: no cover
        args.output = sys.stdout
    args.func(args)



from heyfastqlib.read import (
    Read,
    length,
    qvals,
    trim,
    kscore_ok,
    length_ok,
    trim_moving_average,
    trim_ends,
    seq_id_ok,
)


def test_length():
    assert length(Read("a", "ATCGC", "12345")) == 5


def test_qvals():
    assert qvals(Read("a", "ACGT", "!.:@F")) == [0, 13, 25, 31, 37]


def test_trim():
    a = Read("myseq", "ACGTAC", "123456")
    assert trim(a, end_idx=4) == Read("myseq", "ACGT", "1234")


def test_trim_returns_copy():
    a = Read("a", "ACGG", "FFFF")
    assert length(trim(a, 0, 2)) == 2
    assert length(a) == 4


def test_kscore_ok():
    obs = Read("a", "AAAAC", "12345")  # kscore = 2 / 5 = 0.2
    assert not kscore_ok(obs, min_kscore=0.5)
    assert kscore_ok(obs, min_kscore=0.1)


def test_length_ok():
    obs = Read("g", "ACTTACT", "1234567")
    assert length_ok(obs, 5)
    assert not length_ok(obs, 10)


def test_trim_moving_average():
    r = Read("a", "ACGTACGTAAAAAA", "FFFFFFFF......")
    assert trim_moving_average(r, 3, 15) == Read("a", "ACGTACGT", "FFFFFFFF")


def test_trim_moving_average_endcaps():
    # idx   0  1  2  3 4 5
    # qual 40 40 40 40 0 0
    # ave  40 30 20 10 0 0
    # w25   +  +  -  - - -
    # q25   +  +  +  + - -
    # Window falls below Q25 at idx 2
    # Within window, keep idx 2, 3
    # Trim at 4
    a = Read("a", "ACGTAAAAAAAA", "IIII!!!!!!!!")
    assert trim_moving_average(a, 4, 25) == trim(a, 0, 4)

    # idx   0  1  2  3  4  5 6 7 8 9
    # qual 40 40 40 40  0 40 0 0 0 0
    # ave  40 30 30 20 10 10 0
    # w25   +  +  +  -  -  - -
    # q25   +  +  +  +  -  + - - - -
    # Window falls below Q25 at idx 3
    # Within window, keep 3, 4, 5
    # Trim at 6
    b = Read("b", "CGTTCCCCCCCC", "IIII!I!!!!!!")
    assert trim_moving_average(b, 4, 25) == trim(b, 0, 6)

    # idx   0  1  2  3  4  5  6 7 8 9
    # qual 40 40 40 40  0  0 40 0 0 0
    # ave  40 30 20 20 10 10 10
    # w25   +  +  -  -  -  -  -
    # q25   +  +  +  +  -  -  +
    # Window falls below Q25 at idx 2
    # Within window, keep 2, 3
    # Trim at 4
    c = Read("c", "GCGGACGTCGGG", "IIII!!I!!!!!")
    assert trim_moving_average(c, 4, 25) == trim(c, 0, 4)

    # idx   0  1  2  3  4  5  6 7 8 9
    # qual 40 40 40 40  0  0 40 0 0 0
    # ave  40 30 20 20 10 10 10 0
    # w15   +  +  +  +  -  -  - -
    # q15   +  +  +  +  -  -  + -
    # Window falls below Q15 at idx 4
    # Within window, keep 4, 5, 6
    # Trim at 7
    d = Read("d", "GCGGACGTCGGG", "IIII!!I!!!!!")
    assert trim_moving_average(d, 4, 15) == trim(d, 0, 7)


def test_trim_ends():
    a = Read("bf", "ACGTACGT", "&!FFF!.&")
    # idx  0 1  2  3  4 5  6 7
    # qual 5 0 37 37 37 0 13 5
    # q20  - -  +  +  + -  - - Trim 2:5
    # q6   - -  +  +  + -  + - Trim 2:7
    assert trim_ends(a, 20, 20) == trim(a, 2, 5)
    assert trim_ends(a, 6, 6) == trim(a, 2, 7)


def test_seq_id_ok():
    ids = ["cd", "b"]
    assert not seq_id_ok(Read("b", "GGC", "FFF"), ids)
    assert seq_id_ok(Read("a", "TCG", "FFF"), ids)
    assert seq_id_ok(Read("cd", "TCG", "!!!"), ids, keep=True)
